import numpy as np
np.random.seed(1234)
#np.random.seed(5678)
import scipy.signal

import tensorflow as tf
from keras.backend.tensorflow_backend import set_session
config = tf.ConfigProto()
config.gpu_options.allow_growth = True  # dynamically grow the memory used on the GPU
sess = tf.Session(config=config)
set_session(sess)  # set this TensorFlow session as the default session for Keras
import keras.backend as K
from keras.models import Model
from keras.layers import Input, Conv1D, TimeDistributed
from keras.optimizers import Adam
from keras.constraints import UnitNorm
from keras.callbacks import ModelCheckpoint

from custom_layers import build_forward_stft, TwoSidedToOneSided, Magnitude, identity_filter_init, STFTLayer, ISTFTLayer, Scale, ExpandLastDim, SqueezeLastDim, LastElementOne, UnitNormLastElementOne
from custom_losses import negative_si_mse, si_snr, negative_mag_mse, mag_snr, negative_si_mag_mse, si_mag_snr

class AdversarialFilterModel(object):
    """Optimize a FIR filter to minimize a certain objective

    The FIR filter is applied to an input signal. Using the filtered input
    signal and a reference signal, te specified metric is minimized with respect
    to the filter weights.

    Args:
        metric_to_minimize (str): options are
            mag_snr: SNR in the magnitude STFT domain
            si_mag_snr: scale-invariant SNR in the magnitude STFT domain
            si_sdr: scale-invariant SDR in the time domain
        filter_len (int): length of filter
        stft_window_len (int): duration of STFT window, if using a metric with
            magnitude spectrogram
        unit_norm_constraint (bool): if true, applies a unit-norm constraint to
            the filter during optimization
            
    """
    def __init__(self, num_samples, metric_to_minimize='mag_snr', filter_method='fir', filter_len=512, stft_window_len=512, unit_norm_constraint=False, first_filter_sample_is_one=False, filter_initializer=None, mask_activation='sigmoid', mask_floor=0.05, mask_pnorm=2.0, lr=0.001):

        if not (metric_to_minimize in ['mag_snr', 'si_mag_snr', 'si_sdr']):
            raise ValueError("Unrecognized metric '%s' to minimize" % metric_to_minimize)
        self.metric_to_minimize = metric_to_minimize
        if self.metric_to_minimize == 'si_sdr':
            #self.loss = negative_si_mse
            self.loss = si_snr
            self.metrics = [si_snr]
        elif self.metric_to_minimize == 'mag_snr':
            self.loss = mag_snr
            self.metrics = [mag_snr]
        elif self.metric_to_minimize == 'si_mag_snr':
            self.loss = si_mag_snr
            self.metrics = [si_mag_snr]

        self.filter_method = filter_method
        self.filter = None
        self.filter_len = filter_len
        self.input_shape = (num_samples, 1)
        # STFT parameters
        self.window_len = stft_window_len
        self.window=np.sqrt(scipy.signal.hann(self.window_len, sym=False).astype(np.float32))
        self.hop_len = self.window_len / 4
        self.unit_norm_constraint = unit_norm_constraint
        self.first_filter_sample_is_one = first_filter_sample_is_one
        if self.unit_norm_constraint and self.first_filter_sample_is_one:
            self.filter_constraint = UnitNormLastElementOne()
        elif self.unit_norm_constraint:
            self.filter_constraint = UnitNorm()
        elif self.first_filter_sample_is_one:
            self.filter_constraint = LastElementOne()
        else:
            self.filter_constraint = None
        self.kernel_initializer = filter_initializer
        if self.kernel_initializer is None:
            self.kernel_initializer = 'glorot_uniform'
        elif self.kernel_initializer == 'identity':
            self.kernel_initializer = identity_filter_init
        self.mask_activation = mask_activation
        self.mask_floor = mask_floor
        self.mask_pnorm = mask_pnorm
        self.lr = lr

        self.build_model()


    def build_model(self):
        """Builds and compiles the Keras model used for optimizing the
        adversarial filter
        """
        # input tensor
        x_input = Input(shape=self.input_shape)

        if self.filter_method == 'fir':
          # filtered time-domain input
          x_filt = Conv1D(1, self.filter_len, strides=1, padding='causal', use_bias=False, kernel_constraint=self.filter_constraint, kernel_initializer=self.kernel_initializer)(x_input)

        elif self.filter_method == 'mask':
          x_input_squeezed = SqueezeLastDim(x_input)
          x_stft = STFTLayer(self.window_len, self.hop_len)(x_input_squeezed)
          x_stft_masked = TimeDistributed(Scale(trainable=True, activation=self.mask_activation, floor=self.mask_floor, pnorm=self.mask_pnorm))(x_stft)
          x_filt_squeezed = ISTFTLayer(self.window_len, self.hop_len, samples=self.input_shape[0])(x_stft_masked)
          x_filt = ExpandLastDim(x_filt_squeezed)

        # model with trainable filter weights
        self.model_filt = Model(x_input, x_filt)

        # model that just passed the input through
        self.model_do_nothing = Model(x_input, x_input)

        # compile the models
        self.model_filt.compile(Adam(lr=self.lr), loss=self.loss, metrics=self.metrics)
        self.model_do_nothing.compile(Adam(), loss=self.loss, metrics=self.metrics)


    def fit(self, input_to_filter, reference, gradient_steps=100, verbose=2):
        """Trains the adversarial filter given two signals: a reference signal
        and an input signal that the filter is applied to.

        The reference signal and input_to_filter signals should both be of shape
        (batch_size, num_samples, 1)
        """
        x = input_to_filter
        y = reference
        checkpointer = ModelCheckpoint(filepath='weights.hdf5', verbose=1, save_best_only=True, save_weights_only=True, monitor='val_loss', mode='min')
        self.model_filt.fit(x, y, validation_data=(x, y), epochs=gradient_steps, verbose=verbose, callbacks=[checkpointer])
        self.model_filt.load_weights('weights.hdf5')

        # extract the filter weights from the model
        self.filter = np.squeeze(K.get_value(self.model_filt.trainable_weights[0]))[::-1]


    def evaluate(self, input_to_filter, reference):
        """Evaluates loss and metrics given the current adversarial filter model

        The reference signal and input_to_filter signals should both be of shape
        (batch_size, num_samples, 1)
        """
        x = input_to_filter
        y = reference
        return self.model_filt.evaluate(x, y, verbose=0)


    def apply_filter(self, input_to_filter, mode='same'):
        """Applies the trained adversarial filter to an input signal
        """
        if self.filter is None:
            raise ValueError("No filter exists, first use the fit method to optimize a filter")

        # #implementation of filtering used by mir_eval, except that mir_eval
        # #uses a filtered signal length of nsampl + flen -1 instead of nsampl
        #x = np.squeeze(input_to_filter)
        #return scipy.signal.fftconvolve(x, self.filter[::-1], mode='full')[:len(x)
        if len(input_to_filter.shape) == 1:
            input_to_filter = np.reshape(input_to_filter, (1, len(input_to_filter), 1))
        return self.model_filt.predict(input_to_filter)

