# -*- coding: utf-8 -*-
import numpy as np

import scipy

import tensorflow as tf
from tensorflow.contrib.signal.python.ops import window_ops
from keras import backend as K
from keras import activations, initializers, regularizers, constraints
from keras.models import Model
from keras.engine import Layer, InputSpec
from keras.layers import Input, Dense, Lambda, Conv1D, TimeDistributed, ZeroPadding1D, Multiply
from keras.layers.merge import concatenate
from keras.constraints import Constraint, UnitNorm
from keras.engine.topology import Layer

class LastElementOne(Constraint):
    """Constrains the last element of Conv1D weights to be 1.
    """

    def __call__(self, weight):
        weight_shape = [int(s) for s in weight.shape]
        mask = K.concatenate((K.ones([weight_shape[0]-1] + weight_shape[1:]),
                             K.zeros([1] + weight_shape[1:])), axis=0)
        return mask*weight + (1.0 - mask)

class UnitNormLastElementOne(UnitNorm):
    """Sets last element of weight vector to 1, and the remaining samples are
    constrained to have unit L2 norm.
    """
    def __call__(self, w):
        weight_shape = [int(s) for s in w.shape]
        mask = K.concatenate((K.ones([weight_shape[0]-1] + weight_shape[1:]),
                             K.zeros([1] + weight_shape[1:])), axis=0)
        return (1.0 - mask) + mask*w / (K.epsilon() + K.sqrt(K.sum(K.square(mask*w),
                                               axis=self.axis,
                                               keepdims=True)))


def expand_last_dim(x):
    return K.expand_dims(x, axis=-1)
def output_shape_expand_last_dim(input_shape):
    shape = list(input_shape)
    return tuple(shape + [1])
ExpandLastDim = Lambda(expand_last_dim, output_shape=output_shape_expand_last_dim)

def squeeze_last_dim(x):
    return x[..., 0]
def output_shape_squeeze_last_dim(input_shape):
    shape = list(input_shape)
    return tuple(shape[:-1])
SqueezeLastDim = Lambda(squeeze_last_dim, output_shape=output_shape_squeeze_last_dim)

def identity_filter_init(shape, dtype=None):
    # Assumes filter is of shape (filter_len, 1, 1)
    return K.concatenate([K.zeros((shape[0]-1, 1, 1)), K.ones((1, 1, 1))], axis=0)

def magnitude(x):
    ninput = x.shape[-1]
    return K.sqrt(K.epsilon() + K.square(x[:,:,:ninput/2]) + K.square(x[:,:,ninput/2:]))

def magnitude_squared(x):
    ninput = x.shape[-1]
    return K.square(x[:,:,:ninput/2]) + K.square(x[:,:,ninput/2:])
def magnitude_output_shape(input_shape):
    shape = list(input_shape)
    shape[-1] /= 2
    return tuple(shape)
    
def real_part(x):
    ninput = x.shape[-1]
    return x[:, :, :ninput/2]
def real_part_output_shape(input_shape):
    shape = list(input_shape)
    shape[-1] = shape[-1]/2
    return tuple(shape)

def imag_part(x):
    ninput = x.shape[-1]
    return x[:, :, ninput/2:]
def imag_part_output_shape(input_shape):
    return real_part_output_shape(input_shape)

def tile_last_dim_by_2(x):
    return K.concatenate([x,x], axis=-1)
def tile_last_dim_by_2_output_shape(input_shape):
    shape = list(input_shape)
    shape[-1] = 2*shape[-1]
    return tuple(shape)

def pad_last_dim_by_2(x):
    return K.concatenate([x,K.zeros_like(x)], axis=-1)

def exp_of_j_angle(x):
    x_mag = magnitude(x)
    x_exp_of_j_angle = x/(np.float32(1e-7) + tile_last_dim_by_2(x_mag))
    return x_exp_of_j_angle

def exp_of_j_angle_output_shape(input_shape):
    return input_shape

def same_shape(input_shape):
    return input_shape

def two_sided_to_one_sided(x):
    n = x.shape[-1]/2 # number of redundant frequencies
    return K.concatenate([x[...,:n/2+1], x[..., n:n+n/2+1]], axis=-1)
def two_sided_to_one_sided_output_shape(input_shape):
    shape = list(input_shape)
    n = shape[-1]/2
    shape[-1] = 2*(n/2 + 1)
    return tuple(shape)

MagnitudeSquared = Lambda(magnitude_squared,
    output_shape=magnitude_output_shape)
Magnitude = Lambda(magnitude,
    output_shape=magnitude_output_shape)
Tile2 = Lambda(tile_last_dim_by_2,
               output_shape=tile_last_dim_by_2_output_shape)
Pad2 = Lambda(pad_last_dim_by_2,
              output_shape=tile_last_dim_by_2_output_shape)
ExpOfjPhase = Lambda(exp_of_j_angle,
    output_shape=same_shape)
RealPart = Lambda(real_part,
    output_shape=real_part_output_shape)
ImagPart = Lambda(imag_part,
    output_shape=imag_part_output_shape)
TwoSidedToOneSided = Lambda(two_sided_to_one_sided, output_shape=two_sided_to_one_sided_output_shape)


def build_dft_layer(input_tensor, N, scale='sqrtn'):
    layer_dft = DenseComplex(N, complex_input=False, trainable=False, use_bias=False)
    D = scipy.linalg.dft(N, scale=scale).astype(np.complex64)
    K_dft = layer_dft(input_tensor)
    input_dim = layer_dft.input_shape[-1]
    D = D[:, :input_dim]
    K.set_value(layer_dft.weights[0], np.concatenate([D.real.T, D.imag.T], axis=1) )
    return K_dft, layer_dft


def conjugate(x):
    twice_N = x.shape[-1]
    N = twice_N/2
    x_real = x[..., :N]
    x_imag = x[..., N:]
    return K.concatenate([x_real, -x_imag], axis=-1)

Conjugate = Lambda(conjugate, output_shape=same_shape)

class DenseComplex(Dense):
    #@interfaces.legacy_dense_support
    def __init__(self, units,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 complex_input=False,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(Dense, self).__init__(**kwargs)
        self.units = 2*units
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.input_spec = InputSpec(min_ndim=2)
        self.supports_masking = True
        self.complex_input = complex_input

    def build(self, input_shape):
        assert len(input_shape) >= 2
        input_dim = input_shape[-1]

        """
        self.kernel_real = self.add_weight(shape=(input_dim, self.units/2),
                                      initializer=self.kernel_initializer,
                                      name='kernel_real',
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint)
        self.kernel_imag = self.add_weight(shape=(input_dim, self.units/2),
                                      initializer=self.kernel_initializer,
                                      name='kernel_imag',
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint)
        """
        self.kernel = self.add_weight(shape=(input_dim, self.units),
                                      initializer=self.kernel_initializer,
                                      name='kernel',
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint)

        kernel_real = self.kernel[:, :self.units/2]
        kernel_imag = self.kernel[:, self.units/2:]
        if self.complex_input:
            self.kernel = K.concatenate( [K.concatenate([kernel_real, kernel_imag], axis=1), K.concatenate([-kernel_imag, kernel_real], axis=1)], axis=0)
        else:
            self.kernel = K.concatenate( [kernel_real, kernel_imag], axis=1)

        if self.use_bias:
            self.bias = self.add_weight(shape=(self.units,),
                                        initializer=self.bias_initializer,
                                        name='bias',
                                        regularizer=self.bias_regularizer,
                                        constraint=self.bias_constraint)
        else:
            self.bias = None
        self.input_spec = InputSpec(min_ndim=2, axes={-1: input_dim})
        self.built = True


class Scale(Layer):

    def __init__(self, window=None, complex_input=False, activation='linear', floor=0.0, pnorm=None, **kwargs):
        self.window = window
        self.complex_input = complex_input
        self.activation = activation
        self.floor = floor
        self.pnorm = pnorm
        super(Scale, self).__init__(**kwargs)

    def build(self, input_shape):

        if self.complex_input:
            window_length = input_shape[-1]/2
        else:
            window_length = input_shape[-1]

        if self.window is None:
            # Create a trainable weight variable for this layer.
            self.kernel = self.add_weight(name='kernel', 
                                          shape=(window_length,),
                                          initializer='glorot_normal')
        else:
            self.kernel = self.window
        super(Scale, self).build(input_shape)  # Be sure to call this somewhere!

    def call(self, x):
        if self.activation == 'sigmoid':
            kernel = K.sigmoid(self.kernel)
        else:
            kernel = self.kernel

        if self.floor > 0.0:
          kernel = (1-self.floor)*kernel + self.floor

        if self.pnorm:
          if np.isinf(self.pnorm):
            kernel_pnorm = K.max(K.abs(kernel))
          else:
            kernel_pnorm = K.pow(K.sum(K.pow(kernel, self.pnorm)), 1.0/self.pnorm)
          kernel = kernel / (K.epsilon() + kernel_pnorm)

        if self.complex_input:
            preact = x*tile_last_dim_by_2(kernel)
        else:
            if x.dtype == tf.complex64:
              return tf.complex(kernel*tf.real(x), kernel*tf.imag(x))
            else:
              return x*self.kernel

    def compute_output_shape(self, input_shape):
        return input_shape


def sqrt_hann(frame_length, dtype=None):
  return tf.sqrt(window_ops.hann_window(frame_length, periodic=True))


class STFTLayer(Layer):

    def __init__(self, frame_length, frame_step, **kwargs):
        self.frame_length = frame_length
        self.frame_step = frame_step
        super(STFTLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        super(STFTLayer, self).build(input_shape)  # Be sure to call this somewhere!

    def call(self, x):
        return tf.contrib.signal.stft(x, self.frame_length, self.frame_step,
                                      window_fn=sqrt_hann,
                                      pad_end=True)

    def compute_output_shape(self, input_shape):
        samples = input_shape[-1]
        num_frames = int(4 + np.ceil((samples-self.frame_length)/float(self.frame_step)))
        fft_len = int(pow(2, np.ceil(np.log2(self.frame_length))))
        num_freqs = fft_len // 2 + 1
        return (input_shape[0], num_frames, num_freqs)


class ISTFTLayer(Layer):

    def __init__(self, frame_length, frame_step, samples=None, **kwargs):
        self.frame_length = frame_length
        self.frame_step = frame_step
        self.samples = samples
        super(ISTFTLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        super(ISTFTLayer, self).build(input_shape)  # Be sure to call this somewhere!

    def call(self, x):
        xr = tf.contrib.signal.inverse_stft(x, self.frame_length, self.frame_step,
                                      window_fn=sqrt_hann)
        xr = tf.contrib.signal.inverse_stft(x, self.frame_length, self.frame_step,
                                           window_fn=tf.contrib.signal.inverse_stft_window_fn(self.frame_step, forward_window_fn=sqrt_hann))
        if self.samples is None:
          return xr
        else:
          return xr[..., :self.samples]

    def compute_output_shape(self, input_shape):
        if self.samples is None:
          num_frames = input_shape[1]
          return (input_shape[0], self.frame_length + num_frames*self.frame_step,)
        else:
          return (input_shape[0], self.samples)


def build_layer_forward_frames(winlen, hop, trainable=True, padding="valid"):
    kernel_size = winlen/hop
    kernel_fwd_shape = (kernel_size, hop, winlen)
    kernel_fwd = np.zeros(kernel_fwd_shape, dtype=np.float32)
    kernel_fwd_core = np.concatenate([np.zeros((hop, winlen-hop)), np.eye(hop)], axis=1).astype(np.float32)
    for lag in range(kernel_size):
        kernel_fwd[lag, : ,:] = kernel_fwd_core
        kernel_fwd_core = np.roll(kernel_fwd_core, -hop, axis=1)
    frames = Conv1D(winlen, kernel_size, padding=padding, trainable=trainable)

    return frames, kernel_fwd


def build_forward_stft(x_input, window, hop, Nfft=None, trainable_params=['win', 'dft'], dense_window=False, layers_spec={}, dft_scale='sqrtn', complex_input=False, input_type='nonredundant_frames'):
    """
    Build a forward STFT using Keras.
    Inputs:
    x_input - a Keras tensor that holds nonredundant frames of time-domain,
        of shape (nexamples, time_steps, hop)
    window - a window of shape (winlen,)
    hop - integer that gives window hop in samples
    (trainable_params) - list of trainable parameters in this forward STFT,
        options are 'conv_forward', 'win', 'dft'. Defaults to ['win', 'dft']
    (dense_window) - if True, uses a Dense layer initialized with diag(window)
        to implement windowing. Otherwise, uses a Scale layer to apply the
        window element-wise to the nonredundant frames, which is the
        conventional way windows are applied in the STFT. Defaults to False
    (layers_spec) - a dictionary of specified Keras layers. Used to tie layers.
        Overrides settings of dense_window and trainable_params, using whatever
        the settings of the layers in layers_spec are. Possible keys are 
        'forward_frames', 'apply_window', 'compute_dft'. Defaults to the empty
        dictionary {}
    (dft_scale) - the 'scale' keyword argument passed to scipy.linalg.dft when
        building the DFT matrix
    (complex_input) - whether the time-domain input is complex-valued. If True,
        assumes that the input shape is (nexamples, nframes, 2*hop); otherwise,
        assumes that the input shape is (nexamples, nframes, hop). Defaults to
        False
    (input_type) - options are 'redundant_frames' for input shape of (nexamples, time_steps, hop) or 'raw' for input shape of (nexamples, nsamples, 1)

    Outputs:
    stft_complex - a Keras tensor that corresponds to the complex-valued STFT, with real stacked on top of imaginary
    layers - a dictionary containing the layers of the STFT, with keys ['forward_frames', 'apply_window', 'compute_dft']
    """

    winlen = len(window)
    if input_type == 'nonredundant_frames':
        # build redundant frames from nonredundant frames
        if 'forward_frames' in layers_spec.keys():
            layer_forward_frames = layers_spec['forward_frames']
            if complex_input:
                x_redundant = concatenate([
                    layer_forward_frames(RealPart(x_input)),
                    layer_forward_frames(ImagPart(x_input))
                ], axis=-1)
            else:
                x_redundant = layer_forward_frames(x_input)
        else:
            layer_forward_frames, kernel_forward_frames = build_layer_forward_frames(winlen, hop, trainable=('conv_forward' in trainable_params))
            if complex_input:
                x_redundant = concatenate([
                    layer_forward_frames(RealPart(x_input)),
                    layer_forward_frames(ImagPart(x_input))
                ], axis=-1)
            else:
                x_redundant = layer_forward_frames(x_input)
            K.set_value(layer_forward_frames.weights[0], kernel_forward_frames)
    elif input_type == 'raw':
        kernel = np.reshape(np.eye(winlen).astype(np.float32), (winlen, 1, winlen))
        layer_forward_frames = Conv1D(winlen, winlen, strides=hop, trainable=('conv_forward' in trainable_params))
        x_redundant = layer_forward_frames(x_input)
        K.set_value(layer_forward_frames.weights[0], kernel)

    # apply STFT window to redundant frames
    if 'apply_window' in layers_spec.keys():
        layer_apply_window = layers_spec['apply_window']
        x_windowed = layer_apply_window(x_redundant)
    else:
        if dense_window:
            layer_apply_window = Dense(winlen, trainable=('win' in trainable_params))
            x_windowed = layer_apply_window(x_redundant)
            K.set_value(layer_apply_window.weights[0], np.diag(window))
        else:
            layer_apply_window = TimeDistributed(Scale(trainable=('win' in trainable_params), complex_input=complex_input))
            x_windowed = layer_apply_window(x_redundant)
            K.set_value(layer_apply_window.weights[0], window)

    # compute DFT of windowed frames
    kernel_constraint = None

    if Nfft is None:
        Nfft = winlen

    if 'compute_dft' in layers_spec:
        layer_compute_dft = layers_spec['compute_dft']
        stft_complex = layer_compute_dft(x_windowed)
    else:
        layer_compute_dft = DenseComplex(Nfft, complex_input=complex_input, trainable=('dft' in trainable_params), kernel_constraint=kernel_constraint, use_bias=False)
        D = scipy.linalg.dft(Nfft, scale=dft_scale).astype(np.complex64)
        if Nfft > winlen:
            D = D[:, :winlen]
        stft_complex = layer_compute_dft(x_windowed)
        K.set_value(layer_compute_dft.weights[0], np.concatenate([D.real.T, D.imag.T], axis=1) )

    layers = {'forward_frames': layer_forward_frames,
              'apply_window': layer_apply_window,
              'compute_dft': layer_compute_dft}

    return stft_complex, layers


def compute_forward_stft(x, window=None, hop=None, dft_scale='sqrtn'):
    if window is None:
        WINDOW_LEN = 512
        hop = WINDOW_LEN/4
        window=np.sqrt(scipy.signal.hann(WINDOW_LEN, sym=False).astype(np.float32)) 
    winlen = len(window)

    # enframe and window the input
    kernel = K.variable(np.reshape(np.diag(window), (winlen, 1, winlen)))
    frames = K.conv1d(x, kernel, strides=hop, data_format='channels_last')

    # compute DFT
    dft = scipy.linalg.dft(winlen, scale=dft_scale).astype(np.complex64)
    dft = K.variable(np.concatenate([dft.real.T, dft.imag.T], axis=1))
    return K.dot(frames, dft)

