import os
import numpy as np
import librosa

from metrics import compute_metrics
from adversarial_filter_model import AdversarialFilterModel


def optimize_adversarial_filter(x_enh, x_ref,
    metric_to_minimize='si_sdr',
    filter_method='fir',
    filter_len=512,
    mask_activation='sigmoid',
    mask_floor=0.05,
    mask_pnorm=2.0,
    lr=0.001,
    gradient_steps=100,
    verbose=2,
    unit_norm_constraint=False,
    first_filter_sample_is_one=False,
    filter_initializer=None):
        samples = len(x_ref)
        if len(x_ref) != len(x_enh):
            samples = min(len(x_ref), len(x_enh))
            x_ref = x_ref[:samples]
            x_enh = x_enh[:samples]
        x_ref = np.reshape(x_ref, (1, samples, 1))
        x_enh = np.reshape(x_enh, (1, samples, 1))

        # build the adversarial filter model
        model = AdversarialFilterModel(samples, metric_to_minimize, unit_norm_constraint=unit_norm_constraint, first_filter_sample_is_one=first_filter_sample_is_one, filter_initializer=filter_initializer, filter_method=filter_method, mask_activation=mask_activation, mask_floor=mask_floor, mask_pnorm=mask_pnorm, lr=lr, filter_len=filter_len)

        # build the training inputs and outputs for the model
        x = x_enh
        y = x_ref

        # compute initial metrics
        metrics_initial = compute_metrics(x_ref, x_enh, print_metrics=True, description="Initial")

        loss, metric = model.model_do_nothing.evaluate(x_enh, x_ref, verbose=0)
        print "Initial loss of input is %f" % loss
        print "Initial metric %s of input is %fdB" % (metric_to_minimize, metric)

        x_enh_filt = model.model_filt.predict(x_enh)
        metrics_initial_filt = compute_metrics(x_ref, x_enh_filt, print_metrics=True, description="Initial filtered")

        loss_filt, metric_filt = model.evaluate(x_enh, x_ref)
        print "Initial loss of filtered is %f" % loss_filt
        print "Initial metric %s of filtered is %fdB" % (metric_to_minimize, metric_filt)

        # train the model
        model.fit(x, y, gradient_steps=gradient_steps, verbose=verbose)

        # filter the enhanced signal with the optimized filter, and reevaluate SDR
        x_enh_filt = model.apply_filter(x_enh)
        x_enh_filt = np.reshape(x_enh_filt, (1, samples, 1))
        x = x_enh_filt

        # compute final metrics
        metrics_final_filt = compute_metrics(x_ref, x_enh_filt, print_metrics=True, description="Final filtered")

        loss_final, metric_final = model.evaluate(x_enh, x_ref)
        print "Final loss is %f" % loss_final
        print "Final metric of %s is %fdB" % (metric_to_minimize, metric_final)

        return model, metrics_initial, metrics_initial_filt, metrics_final_filt


if __name__=="__main__":
    """
    Find a 512-length FIR filter that minimizes SNR in the magnitude spectrogram domain
    """
    wav_ref = 'chime2-wsj0/scaled/si_et_05/0dB/wsj0/si_et_05/440/440c0201.wav'
    x_ref, fs = librosa.load(wav_ref, sr=16000)
    x_enh = x_ref

    filter_len = 512
    lr = 0.001
    gradient_steps = 400
    metric_to_minimize = 'si_sdr'

    unit_norm_constraint = False

    first_filter_sample_is_one = False

    filter_initializer = 'identity'

    mask_activation = None
    mask_floor = 0.0
    mask_pnorm = 2.0
    filter_method = 'mask'; lr=0.1; mask_activation='sigmoid'; mask_floor=0.0; mask_pnorm=np.inf

    #verbose = 2
    verbose = 0

    # set the key of the corresponding metric within the dictionary returned
    # by compute_metrics
    if metric_to_minimize == 'si_sdr':
        metric_key = 'SI-SDR'
    elif metric_to_minimize == 'mag_snr':
        metric_key = 'MagSNR'
    elif metric_to_minimize == 'si_mag_snr':
        metric_key = 'SI-MagSNR'

    seed = 47629
    num_trials = 1
    best_sdr_minus_metric = -np.Inf
    for i_trial in range(num_trials):
      seed = seed + i_trial*1031
      np.random.seed(seed)

      model, metrics_initial, metrics_initial_filt, metrics_final_filt = optimize_adversarial_filter(x_enh, x_ref, metric_to_minimize=metric_to_minimize, gradient_steps=gradient_steps, verbose=verbose, unit_norm_constraint=unit_norm_constraint, first_filter_sample_is_one=first_filter_sample_is_one, filter_initializer=filter_initializer, filter_method=filter_method, mask_activation=mask_activation, mask_floor=mask_floor, mask_pnorm=mask_pnorm, lr=lr, filter_len=filter_len)

      print 'For seed %d, filtered enhanced, SDR = %.2fdB, %s = %.2fdB' % (seed, metrics_final_filt['SDR'], metric_key, metrics_final_filt[metric_key])

      sdr = metrics_final_filt['SDR']
      metric = metrics_final_filt[metric_key]
      if (sdr - metric) > best_sdr_minus_metric:
          best_sdr_minus_metric = sdr - metric
          print "New best (SDR - %s)! SDR = %.2fdB, %s = %.2fdB, diff = %.2fdB" % (metric_key, sdr, metric_key, metric, sdr-metric)

      print "End of trial %d of %d" % (i_trial+1, num_trials)
      print ""
      print ""

    if filter_method == 'fir':
        description = "%s_%s_flen%d_lr%.3f_%dsteps" % (metric_to_minimize, filter_method, filter_len, lr, gradient_steps)
    elif filter_method == 'mask':
        description = "%s_%s_act-%s_lr%.3f_%dsteps_l%.0fnorm" % (metric_to_minimize, filter_method, mask_activation, lr, gradient_steps, mask_pnorm)

    filt = model.filter
    x_enh_filt = model.apply_filter(x_enh)
    if not os.path.exists('audio'):
      os.makedirs('audio')
    librosa.output.write_wav('audio/enh_filt_%s.wav' % description, np.squeeze(x_enh_filt), 16000)
    librosa.output.write_wav('audio/ref_%s.wav' % description, np.squeeze(x_ref), 16000)


    import matplotlib as mpl
    mpl.use('Agg')
    import matplotlib.pyplot as plt
    from scipy.signal import freqz

    if not os.path.exists('figures'):
      os.makedirs('figures')

    if filter_method == 'fir':
      # Plot the frequency response of the filter
      plt.figure(1)
      plt.clf()
      w, h = freqz(filt[::-1], 1, worN=2000)
      plt.plot((fs * 0.5 / np.pi) * w, 20.*np.log10(abs(h)))

      plt.xlabel('Frequency (Hz)')
      plt.ylabel('Gain')
      plt.grid(True)
      plt.title('Frequency response of filter minimizing %s' % metric_to_minimize)
      plt.savefig('filter.pdf' % description, bbox_inches='tight')

      plt.figure(1)
      plt.clf()
      plt.plot(filt)
      plt.xlabel('Sample index')
      plt.grid(True)
      plt.title('Impulse response of filter minimizing %s' % metric_to_minimize)
      plt.savefig('adversarial_filter_impulse_%s.pdf' % description, bbox_inches='tight')

    elif filter_method == 'mask':
      # Plot the frequency response of the filter
      for figsize, suffix in zip([[8.0, 6.0], [8.0, 4.0]], ['', '_poster']):
        plt.figure(figsize=figsize)
        plt.clf()
        mask = mask_floor + (1.0 - mask_floor)/(1+np.exp(-filt[::-1]))
        mask = mask / (1e-7 + pow(np.sum(pow(mask, mask_pnorm)), 1.0/mask_pnorm))
        p1 = plt.plot(mask, label='Mask minimizing SI-SDR')

        plt.xlabel('Frequency bin')
        plt.grid(True)
        plt.title('Frequency responses of filters')
        ax1 = plt.gca()
        ax1.set_ylabel('Gain', color='b')
        ax1.tick_params('y', colors=[0.0, 0.0, 1.0])

        sdr_filters = metrics_final_filt['SDR_filters']
        w, h = freqz(sdr_filters[:, 0], 1, worN=2000)
        ax2 = ax1.twinx()
        p2 = ax2.plot(np.linspace(0, len(mask), len(h)), 20.*np.log10(abs(h)), color='r', label='Filter estimated by SDR')
        ax2.set_ylabel('Gain (dB)', color='r')
        ax2.tick_params('y', colors='r')
        plt.legend((p1[0], p2[0]), ('Mask minimizing SI-SDR', 'Filter estimated by SDR'))

        plt.savefig('figures/filter%s.pdf' % suffix, bbox_inches='tight')

    # Plot spectrograms
    win_len = 1024
    X_ref = np.abs(librosa.stft(np.squeeze(x_ref), n_fft=win_len, hop_length=win_len/4, win_length=win_len))
    X_enh = np.abs(librosa.stft(np.squeeze(x_enh), n_fft=win_len, hop_length=win_len // 4, win_length=win_len))
    X_enh_filt = np.abs(librosa.stft(np.squeeze(x_enh_filt), n_fft=win_len, hop_length=win_len // 4, win_length=win_len))

    fig, ax = plt.subplots(2, 1, figsize=(20,10))
    ax1, ax2 = ax
    ims = []
    vamin=np.Inf
    vamax=-np.Inf
    # spectrogram of reference signal
    sgram=20.*np.log10(1e-7+np.squeeze(X_ref))
    vmin = np.min(sgram)
    if vmin < vamin:
        vamin=vmin
    vmax = np.max(sgram)
    if vmax > vamax:
        vamax = vmax
    im1=ax1.imshow(sgram, aspect='auto', interpolation=None)
    ims.append(im1)
    ax1.invert_yaxis()
    metric_to_minimize = metric_to_minimize.upper()
    metric_to_minimize = metric_to_minimize.replace('_', '-')
    ax1.set_title('Reference, SDR = %.2fdB, SNR = %.2fdB, %s = %.2fdB' % (metrics_initial['SDR'], metrics_initial['SNR'], metric_to_minimize, metrics_initial[metric_key]))
    cbar = fig.colorbar(im1, ax=ax1)
    ticklabs = cbar.ax.get_yticklabels()
    cbar.ax.set_yticklabels(ticklabs, fontsize=16)

    # spectrogram of filtered signal
    ax3 = ax2
    sgram=20.*np.log10(1e-7+np.squeeze(X_enh_filt))
    vmin = np.min(sgram)
    if vmin < vamin:
        vamin=vmin
    vmax = np.max(sgram)
    if vmax > vamax:
        vamax = vmax
    im3=ax3.imshow(sgram, aspect='auto', interpolation=None)
    ims.append(im3)
    ax3.invert_yaxis()
    metric_to_minimize = metric_to_minimize.upper()
    metric_to_minimize = metric_to_minimize.replace('_', '-')
    ax3.set_title('Filtered, SDR = %.2fdB, SNR = %.2fdB, %s = %.2fdB' % (metrics_final_filt['SDR'], metrics_final_filt['SNR'], metric_to_minimize, metrics_final_filt[metric_key]))
    cbar = fig.colorbar(im3, ax=ax3)
    ticklabs = cbar.ax.get_yticklabels()
    cbar.ax.set_yticklabels(ticklabs, fontsize=16)

    for im in ims:
        im.set_clim(vamax-80, vamax)

    for ax in [ax1, ax2]:
      for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()):
        item.set_fontsize(16)

    plt.savefig('figures/spectrograms.pdf', bbox_inches='tight')
