import numpy as np
import mir_eval
import keras.backend as K

import custom_mir_eval
from custom_layers import compute_forward_stft, TwoSidedToOneSided, Magnitude


def compute_metrics(x_ref, x_enh, print_metrics=False, description=''):
    """ Compute a set of source separation metrics for a reference signal
    ref and an enhanced signal enh. The inputs ref and enh should be of
    shape (1, num_samples, 1).

    Optionally, the metrics can be printed if print_metrics=True, using the
    string description as:
    '<description> <metric> is <value> dB'

    Returns a dictionary with the metrics
    """
    metrics = {}

    # SDR
    sdr, sir, sar, perm, C_spat, C_interf = custom_mir_eval.bss_eval_sources(
        x_ref[:, :, 0],
        x_enh[:, :, 0],
        compute_permutation=False,
        return_filter=True)
    metrics['SDR'] = sdr
    metrics['SDR_filters'] = C_spat

    # Time-domain SNR
    snr_time = 10.*np.log10(np.mean(x_ref**2)/np.mean((x_ref-x_enh)**2))
    metrics['SNR'] = snr_time

    # Scale-invariant SDR (SI-SDR)
    si_sdr = compute_si_sdr(np.squeeze(x_ref), np.squeeze(x_enh))
    metrics['SI-SDR'] = si_sdr

    # Magnitude SDR
    mag_snr = compute_mag_snr(x_ref, x_enh)
    metrics['MagSNR'] = mag_snr

    # Scale-invariant magnitude SDR
    si_mag_snr = compute_si_mag_snr(x_ref, x_enh)
    metrics['SI-MagSNR'] = si_mag_snr

    # Scale-dependent SDR (SI-SDR)
    sd_sdr = compute_sd_sdr(np.squeeze(x_ref), np.squeeze(x_enh))
    metrics['SD-SDR'] = sd_sdr

    if print_metrics:
        print ""
        for key, value in metrics.iteritems():
            if not ('filters' in key):
              print "%s %s is %.2f dB" % (description, key, value)

    return metrics


def compute_spectrogram(x):
    x_K = K.placeholder(shape=x.shape)
    x_stft_K = compute_forward_stft(x_K)
    x_stft1_K = TwoSidedToOneSided(x_stft_K)
    x_mag_K = Magnitude(x_stft1_K)
    x_mag = x_mag_K.eval(feed_dict={x_K:x}, session=K.get_session())
    return x_mag


def compute_snr(ref, enh):
    return 10.*np.log10( np.sum(ref**2) / np.sum( (ref-enh)**2 ) )


def compute_si_sdr(ref, enh):
    scale = np.sum(ref * enh) / (np.sum(ref**2))
    ref_scaled = scale*ref
    return compute_snr(ref_scaled, enh)


def compute_sd_sdr(ref, enh):
    scale = np.sum(ref * enh) / (np.sum(ref**2))
    ref_scaled = scale*ref
    return 10*np.log10( np.sum(ref_scaled**2) / np.sum( (ref-enh)**2 ) )


def compute_mag_snr(ref, enh):
    # compute ref mag stft
    ref_mag = compute_spectrogram(ref)

    # compute enh mag stft
    enh_mag = compute_spectrogram(enh)

    return compute_snr(ref_mag, enh_mag)


def compute_si_mag_snr(ref, enh):
    scale = np.sum(np.squeeze(ref) * np.squeeze(enh)) / (np.sum(ref**2))
    ref_scaled = scale*ref
    return compute_mag_snr(ref_scaled, enh)

