import os
import numpy as np
import librosa
import scipy.signal
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import hickle

from metrics import compute_spectrogram, compute_metrics
import custom_librosa


def mask_from_filt_spec(freqs, freq_spec, gain_spec):
  gains = np.zeros((len(freqs),))
  freq_prev = 0.
  for freq, gain in zip(freq_spec, gain_spec):
    band = (freqs >= freq_prev) & (freqs < freq)
    freq_prev = freq
    gains[band] = gain
  top_band = freqs >= freq_spec[-1]
  gains[top_band] = gain_spec[-1]
  return gains


def compute_stft(x, n_win=512, n_fft=1024, hop_div=4):
  x = np.squeeze(x)
  n = len(x)
  x_pad = np.pad(x, (n_win, n_win), 'constant', constant_values=(0., 0.))
  window = np.sqrt(scipy.signal.hann(n_win, sym=False).astype(np.float32))
  D = librosa.stft(x_pad, n_fft=n_fft, win_length=n_win, window=window, hop_length=n_win/hop_div)
  return D


def filter_with_mask(x, mask, n_win=512, n_fft=1024, hop_div=4):
  x = np.squeeze(x)
  n = len(x)
  window = np.sqrt(scipy.signal.hann(n_win, sym=False).astype(np.float32))
  D = compute_stft(x, n_win=n_win, n_fft=n_fft, hop_div=hop_div)
  D_masked = D*mask[:, np.newaxis]
  xr = custom_librosa.istft_nodiv(D_masked, length=n+2*n_win, win_length=n_win, window=window/(hop_div/2), hop_length=n_win/4)[n_win:-n_win]
  return xr, D, D_masked


if __name__=="__main__":
    """
    Successively design a 512-length filter to mask out more and more frequencies, then plot metrics against how many frequency bins are masked
    """
    wav_ref = 'chime2-wsj0/scaled/si_et_05/0dB/wsj0/si_et_05/440/440c0201.wav'
    x_ref, fs = librosa.load(wav_ref, sr=16000)
    samples = len(x_ref)
    x_ref = np.reshape(x_ref, (1, samples, 1))

    # Set parameters
    noise_type = 'bandpass_white'

    filter_method = 'mask_stft'

    filter_len = 511 # use odd filter length for type I

    # STFT parameters
    n_win = 512
    n_fft = 1024
    hop_div = 4
    f_nyquist = fs/2
    freqs = np.linspace(0, f_nyquist, n_fft/2 + 1) / f_nyquist
    plot_spectrograms = False

    snrs_band = [0]
    proportions = [0.1]
    # set the gain step
    step = 0.025
    # create the list of bandstop gains
    filt_gains_bandstop = np.arange(1., -step, -step)

    # seed for the noise
    rng_seed = 1234

    freq_center = 'max_spec'
    if freq_center == 'max_spec':
      def find_max_freq(x_enh):
          # find the maximum frequency bin
          x_enh_mag = compute_spectrogram(x_enh)
          num_freqs = x_enh_mag.shape[-1]
          mean_spectrum = np.mean(x_enh_mag[0, :, :], axis=0)
          idx_max_freq = np.argmax(mean_spectrum)
          return float(idx_max_freq)/num_freqs
    elif freq_center == 'middle':
      def find_max_freq(x_enh):
        return 0.5
    else:
      raise ValueError("Unknown freq_center of '%s'" % freq_center)

    suffix = '%s_fcenter-%s' % (filter_method, freq_center)
    savefile_name = 'metrics_%s.hdf5' % suffix

    sdrs_all = {snr: {prop: None for prop in proportions} for snr in snrs_band}
    si_sdrs_all = {snr: {prop: None for prop in proportions} for snr in snrs_band}
    sd_sdrs_all = {snr: {prop: None for prop in proportions} for snr in snrs_band}
    snrs_all = {snr: {prop: None for prop in proportions} for snr in snrs_band}
    mag_snrs_all = {snr: {prop: None for prop in proportions} for snr in snrs_band}

    flag_recompute = True
    if (not os.path.exists(savefile_name)) or flag_recompute:

      for snr in snrs_band:
        for proportion in proportions:

          print "Generating data for SNR %ddB, bandwidth %.3f" % (snr, proportion)

          if noise_type == 'bandpass_white':
              x_enh = x_ref

              max_freq = find_max_freq(x_enh)

              # create some bandpass noise
              lower_freq = max(0., max_freq - (proportion*max_freq))
              upper_freq = min(1., max_freq + (proportion*(1.-max_freq)))
              freq = [0., lower_freq, lower_freq, upper_freq, upper_freq, 1.0]
              gain = [0., 0., 1., 1., 0., 0.]
              filt = scipy.signal.firwin2(filter_len, freq, gain)
              rng = np.random.RandomState(1234)
              noise = rng.randn(*x_ref.shape)
              noise_bp = scipy.signal.fftconvolve(np.squeeze(noise), filt)[:samples]
              noise_bp = np.reshape(noise_bp, (1, samples, 1))
              x_ref_bp = scipy.signal.fftconvolve(np.squeeze(x_ref), filt)[:samples]
              x_ref_bp = np.reshape(x_ref_bp, (1, samples, 1))
              scale = np.sqrt(10**(-snr/10.)*np.sum(x_ref_bp**2)/np.sum(noise_bp**2))
              x_enh = x_ref + scale*noise_bp

              # check SNR
              snr_check = 10*np.log10( np.sum(x_ref_bp**2) / (np.sum( (scale*noise_bp)**2 )) )

              if plot_spectrograms:
                librosa.output.write_wav('mix.wav', np.squeeze(x_enh), 16000)
              if plot_spectrograms:
                D = compute_stft(x_ref)
                plt.figure()
                plt.imshow(20*np.log10(np.abs(D)))
                plt.colorbar()
                plt.savefig('spectrogram_ref.png')

          elif noise_type == 'chime2':
              wav_enh = 'chime2-wsj0/isolated/si_et_05/0dB/wsj0/si_et_05/440/440c0201.wav'
              x_enh, fs = librosa.load(wav_enh, sr=16000)
              x_enh = np.reshape(x_enh, (1, samples, 1))

              max_freq = find_max_freq(x_enh)

          # initialize lists to store the metrics
          sdrs = []
          si_sdrs = []
          sd_sdrs = []
          snrs = []
          mag_snrs = []

          ## evaluate initial metrics

          for filt_gain in filt_gains_bandstop:
              # bandstop filter on noise

              # set the filter freqs and gains
              lower_freq = max_freq - (proportion*max_freq)
              upper_freq = max_freq + (proportion*(1.-max_freq))
              freq = [0., lower_freq, lower_freq, upper_freq, upper_freq, 1.0]
              gain = [1., 1., filt_gain, filt_gain, 1., 1.]

              if filter_method == 'fir_fftconv':
                # use window design to create a FIR filter with desired response
                filt = scipy.signal.firwin2(filter_len, freq, gain)

                # apply the filter
                x_enh_filt = scipy.signal.fftconvolve(np.squeeze(x_enh), filt)[:samples]

                if plot_spectrograms:
                  # plot the frequency response
                  w, response = scipy.signal.freqz(filt)
                  plt.figure()
                  plt.plot(w, 20 * np.log10(abs(response)))
                  plt.ylim(-140, 3)
                  plt.grid()
                  plt.savefig('figures/response_%.3fgain_proportion%.3f_bandstop.png' % (filt_gain, proportion))

              elif filter_method == 'mask_stft':
                # apply ideal filter in the STFT domain as a mask
                mask = mask_from_filt_spec(freqs, freq, gain)
                x_enh_filt, D, D_masked = filter_with_mask(np.squeeze(x_enh), mask, n_win=n_win, n_fft=n_fft)
                if plot_spectrograms:
                  plt.figure()
                  plt.imshow(20*np.log10(np.abs(D)))
                  plt.colorbar()
                  plt.savefig('spectrogram_mix.png')
                  plt.figure()
                  plt.imshow(20*np.log10(np.abs(D_masked)))
                  plt.colorbar()
                  plt.savefig('spectrogram_enh_%.3fgain_%s.png' % (filt_gain, filter_method))

              else:
                raise ValueError("Unknown filter_method of '%s'" % filter_method)

              if plot_spectrograms:
                librosa.output.write_wav('enh_%.3fgain_%s.wav' % (filt_gain, filter_method), np.squeeze(x_enh_filt), 16000)

              x_enh_filt = np.reshape(x_enh_filt, (1, samples, 1))

              # compute metrics
              metrics = compute_metrics(x_ref, x_enh_filt, print_metrics=True, description="Notch band gain of %.3f:" % filt_gain)
              sdrs.append(metrics['SDR'])
              si_sdrs.append(metrics['SI-SDR'])
              sd_sdrs.append(metrics['SD-SDR'])
              snrs.append(metrics['SNR'])
              mag_snrs.append(metrics['MagSNR'])

          sdrs_all[snr][proportion] = sdrs
          si_sdrs_all[snr][proportion] = si_sdrs
          sd_sdrs_all[snr][proportion] = sd_sdrs
          snrs_all[snr][proportion] = snrs
          mag_snrs_all[snr][proportion] = mag_snrs

          hickle.dump({'SDR': sdrs_all, 'SI-SDR': si_sdrs_all, 'SD-SDR': sd_sdrs_all, 'SNR': snrs_all, 'MagSNR': mag_snrs_all}, savefile_name)

    else:
      L = hickle.load(savefile_name)
      sdrs_all = L['SDR']
      si_sdrs_all = L['SI-SDR']
      sd_sdrs_all = L['SD-SDR']
      snrs_all = L['SNR']
      mag_snrs_all = L['MagSNR']

    x = filt_gains_bandstop
    if len(snrs_band) > 1 and len(proportions) > 1:
      fig, ax = plt.subplots(len(snrs_band), len(proportions), sharey=True, figsize=(15,15))
    else:
      plt.figure()
    for i_snr, snr in enumerate(snrs_band):
      for i_prop, proportion in enumerate(proportions):
        if len(snrs_band) > 1 and len(proportions) > 1:
          plt.sca(ax[i_snr][i_prop])
        sdrs = sdrs_all[snr][proportion]
        si_sdrs = si_sdrs_all[snr][proportion]
        sd_sdrs = sd_sdrs_all[snr][proportion]
        snrs = snrs_all[snr][proportion]
        mag_snrs = mag_snrs_all[snr][proportion]
        plt.plot(x, sdrs, label='SDR (global)')
        plt.plot(x, snrs, label='SNR (global)')
        plt.plot(x, si_sdrs, label='SI-SDR (global)')
        plt.plot(x, sd_sdrs, label='SD-SDR (global)')
        ylim = plt.ylim()
        plt.xlabel("Noise band gain")
        plt.ylabel("dB")
        plt.grid()
    if len(snrs_band) > 1 and len(proportions) > 1:
      plt.sca(ax[0][0])
    plt.legend()
    plt.savefig('figures/metrics_vs_gain_subplots_%s_withsdsdr.pdf' % suffix, bbox_inches='tight')
