import numpy as np
import librosa
import scipy.signal
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

from metrics import compute_spectrogram, compute_metrics
import custom_librosa


def mask_from_filt_spec(freqs, freq_spec, gain_spec):
  gains = np.zeros((len(freqs),))
  freq_prev = 0.
  for freq, gain in zip(freq_spec, gain_spec):
    band = (freqs >= freq_prev) & (freqs < freq)
    freq_prev = freq
    gains[band] = gain
  top_band = freqs >= freq_spec[-1]
  gains[top_band] = gain_spec[-1]
  return gains


def compute_stft(x, n_win=512, n_fft=1024, hop_div=4):
  x = np.squeeze(x)
  n = len(x)
  x_pad = np.pad(x, (n_win, n_win), 'constant', constant_values=(0., 0.))
  window = np.sqrt(scipy.signal.hann(n_win, sym=False).astype(np.float32))
  D = librosa.stft(x_pad, n_fft=n_fft, win_length=n_win, window=window, hop_length=n_win/hop_div)
  return D


def filter_with_mask(x, mask, n_win=512, n_fft=1024, hop_div=4):
  x = np.squeeze(x)
  n = len(x)
  window = np.sqrt(scipy.signal.hann(n_win, sym=False).astype(np.float32))
  D = compute_stft(x, n_win=n_win, n_fft=n_fft, hop_div=hop_div)
  D_masked = D*mask[:, np.newaxis]
  xr = custom_librosa.istft_nodiv(D_masked, length=n+2*n_win, win_length=n_win, window=window/(hop_div/2), hop_length=n_win/4)[n_win:-n_win]
  return xr, D, D_masked

if __name__=="__main__":
    """
    Successively design a 512-length filter to mask out more and more frequencies, then plot metrics against how many frequency bins are masked
    """
    wav_ref = 'chime2-wsj0/scaled/si_et_05/0dB/wsj0/si_et_05/440/440c0201.wav'
    wav_enh = 'chime2-wsj0/scaled/si_et_05/0dB/wsj0/si_et_05/440/440c0201.wav'
    x_ref, fs = librosa.load(wav_ref, sr=16000)
    x_enh, fs = librosa.load(wav_enh, sr=16000)

    samples = len(x_ref)
    if len(x_ref) != len(x_enh):
        samples = min(len(x_ref), len(x_enh))
        x_ref = x_ref[:samples]
        x_enh = x_enh[:samples]
    x_ref = np.reshape(x_ref, (1, samples, 1))

    # controls whether to use a noisy or clean signal
    signal_type = 'noisy'; snr = 15
    x_enh = np.reshape(x_enh, (1, samples, 1))

    filter_method = 'mask_stft'

    filter_len = 511 # use odd filter length for type I

    which_freq = 'median'

    suffix = "%s_whichfreq-%s" % (filter_method, which_freq)

    # STFT parameters
    n_win = 512
    n_fft = 1024
    hop_div = 4
    f_nyquist = fs/2
    freqs = np.linspace(0, f_nyquist, n_fft/2 + 1) / f_nyquist
    plot_spectrograms = False

    sdrs = []
    si_sdrs = []
    snrs = []
    mag_snrs = []
    sd_sdrs = []

    if signal_type == 'clean':
        x_enh = x_ref
    elif signal_type == 'noisy':
        rng = np.random.RandomState(1234)
        noise = rng.randn(*x_ref.shape)
        scale = np.sqrt(10**(-snr/10.)*np.sum(x_ref**2)/np.sum(noise**2))
        x_enh = x_ref + scale*noise
        suffix += ('_snr-%ddB' % snr)

    # evaluate initial metrics
    metrics = compute_metrics(x_ref, x_enh, print_metrics=True, description="Unfiltered")
    sdrs.append(metrics['SDR'])
    si_sdrs.append(metrics['SI-SDR'])
    snrs.append(metrics['SNR'])
    mag_snrs.append(metrics['MagSNR'])
    sd_sdrs.append(metrics['SD-SDR'])

    # find the maximum frequency bin
    x_enh_mag = compute_spectrogram(x_enh)
    num_freqs = x_enh_mag.shape[-1]
    mean_spectrum = np.mean(x_enh_mag[0, :, 1:], axis=0)
    if which_freq == 'max':
      idx_max_freq = np.argmax(mean_spectrum)
    elif which_freq == 'min':
      idx_max_freq = np.argmin(mean_spectrum)
    elif which_freq == 'median':
      idx_max_freq = np.argsort(mean_spectrum)[len(mean_spectrum)//2]
    max_freq = float(idx_max_freq)/num_freqs
    print "Starting freq is %.3f" % max_freq

    step = 0.025
    proportions_to_mask = np.arange(step, 1., step)
    for proportion in proportions_to_mask:

        # create the filter
        lower_freq = max_freq - (proportion*max_freq)
        upper_freq = max_freq + (proportion*(1.-max_freq))
        freq = [0., lower_freq, lower_freq, upper_freq, upper_freq, 1.0]
        gain = [1., 1., 0., 0., 1., 1.]

        if filter_method == 'fir_fftconv':
          filt = scipy.signal.firwin2(filter_len, freq, gain)

          # apply the filter
          x_enh_filt = scipy.signal.fftconvolve(np.squeeze(x_enh), filt)[:samples]
          x_enh_filt = np.reshape(x_enh_filt, (1, samples, 1))

          if plot_spectrograms:
            # plot the frequency response
            w, response = scipy.signal.freqz(filt)
            plt.figure()
            plt.plot(w, 20 * np.log10(abs(response)))
            plt.ylim(-140, 3)
            plt.grid()
            plt.savefig('plots/response_%.3f_masked.png' % proportion)

        elif filter_method == 'mask_stft':
          # apply ideal filter in the STFT domain as a mask
          mask = mask_from_filt_spec(freqs, freq, gain)
          x_enh_filt, D, D_masked = filter_with_mask(np.squeeze(x_enh), mask, n_win=n_win, n_fft=n_fft)
          x_enh_filt = np.reshape(x_enh_filt, (1, samples, 1))
          if plot_spectrograms:
            plt.figure()
            plt.imshow(20*np.log10(np.abs(D)))
            plt.colorbar()
            plt.savefig('spectrogram_mix.png')
            plt.figure()
            plt.imshow(20*np.log10(np.abs(D_masked)))
            plt.colorbar()
            plt.savefig('spectrogram_enh_%.3fgain_%s.png' % (filt_gain, filter_method))

        else:
          raise ValueError("Unknown filter_method of '%s'" % filter_method)

        # compute metrics
        metrics = compute_metrics(x_ref, x_enh_filt, print_metrics=True, description="Proportion of %.3f" % proportion)
        sdrs.append(metrics['SDR'])
        si_sdrs.append(metrics['SI-SDR'])
        snrs.append(metrics['SNR'])
        mag_snrs.append(metrics['MagSNR'])
        sd_sdrs.append(metrics['SD-SDR'])

    proportions_to_mask = np.arange(0., 1., step)
    plt.figure()
    plt.plot(proportions_to_mask, sdrs, label='SDR')
    plt.plot(proportions_to_mask, snrs, label='SNR')
    plt.plot(proportions_to_mask, si_sdrs, label='SI-SDR')
    plt.plot(proportions_to_mask, sd_sdrs, label='SD-SDR')
    ylim = plt.ylim()
    if filter_method == 'mask_stft':
      if signal_type == 'clean':
        plt.ylim(max(-30, np.floor(si_sdrs[-2])), np.ceil(sdrs[1]))
      elif signal_type == 'noisy' and which_freq=='median':
        plt.ylim(max(-30, np.floor(si_sdrs[-2])), 17)
    elif filter_method == 'fir_fftconv':
      plt.ylim(ylim[0], 1.05*np.max(sdrs[1:]))
    plt.xlabel('Proportion of bins masked')
    plt.ylabel('dB')
    plt.grid()
    plt.legend()
    plt.savefig('figures/metrics_vs_deleted_bins_withsdsdr.pdf', bbox_inches='tight')

